// ----------------------------------------------------------------
// Bazas
//     v. 1.2.0
//     Copyright 2003 Gonzalo Mena-Mendoza
//     gonzalo@mena.com.mx
//     http://mena.com.mx
// ----------------------------------------------------------------
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by he Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied
// warranty of MERCHANTABILITY or FITNESS FOR A
// PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free
// Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA  02111-1307  USA
// ----------------------------------------------------------------
#include "os_35_extra.h"
#define AppCreator 'Baza'
#define MinROMVersion 0x03503000
#define VersionRes 1

#define MainForm 1000
#define PlayerForm 1001
#define WantForm 1002
#define GotForm 1003
#define StatsForm 1004
#define MainMenuBar 1000
#define NewMenuOpt 1000
#define StatsMenuOpt 1001
#define AboutMenuOpt 1002
#define RulesBaseMenuOpt 1100
#define RulesBaseStr 1100
#define NewGameAlert 1000
#define MinPlayersAlert 1001
#define DealAlert 1002
#define AboutAlert 1003
#define DebugAlert 1099
#define IncompatAlert 1100
#define PlayerTable 1001
#define HandScrollBar 1002
#define BackButton 1100
#define NextButton 1101
#define RoundsNameLabel 1001
#define RoundsAcumLabel 1003
#define RoundsWantLabel 1003
#define RoundsBaseButton 1010
#define StatsHandsLabel 1001

#define MinPlayers 3
#define MaxPlayers 7
#define MinRounds MinPlayers
#define MaxRounds MaxPlayers
#define MaxHands (MaxPlayers + (MaxRounds-MinRounds)*2)

#define MaxRules 6
#define MaxNameLen 8
#define VisibHands 6

// hands grid size
#define TX 0
#define TY 16
#define PH 12
#define PW 20
#define HH 22
#define HW 12

// stats grid size
#define ST_TX 4
#define ST_TY 14
#define ST_NW 50
#define ST_SW 24
#define ST_RH 12

#define EMPTY -1

typedef enum {
    SET = 0,
    WANT,
    PLAY,
    GOT,
    REST
} StageEnum;

typedef struct {
    int want;
    int got;
} Hand;

typedef struct {
    char players[MaxPlayers][MaxNameLen+1];
    Hand hands[MaxHands][MaxPlayers];
    int startRounds;
} Prefs;
Prefs prefs;

char *versionP = "1.x.y";
Handle nameHandles[MaxPlayers];
int pW;
RGBColorType rowBackEven =
    {0x00, 0xff, 0xff, 0xff};
RGBColorType rowBackOdd =
    {0x00, 0xcc, 0xcc, 0xcc};

int roundsPerHand[MaxHands];
int points[MaxHands][MaxPlayers];
int numPlayers;
int numHands;
int firstVisibHand = 0;
int currHand;
int currPlayer;
StageEnum stage;

static int actualPlayer(int hand, int player) {
    return (
        (player + hand + numPlayers)
        % numPlayers);
}

static void updateHands() {
    // calculate rounds/hand
    int h = 0;
    for (int i=prefs.startRounds; i < MaxRounds; ++i)
        roundsPerHand[h++] = i;
    for (int i=0; i < numPlayers; ++i)
        roundsPerHand[h++] = MaxRounds;
    for (int i=MaxRounds-1; i >=prefs.startRounds; --i)
        roundsPerHand[h++] = i;
    // total of hands
    numHands = h;
}

static void clearAllPoints() {
    for (int h=0; h<MaxHands; ++h)
        for (int p=0; p<MaxPlayers; ++p) {
            prefs.hands[h][p].want = EMPTY;
            prefs.hands[h][p].got = EMPTY;
            points[h][p] = EMPTY;
        }
}

static int calcNewPoints(int want, int got) {
    if ((EMPTY == want) || (EMPTY == got))
        return EMPTY;
    if (got == want)
        return (10 + 5*want);
    else
        return (-5*((want>got)?(want-got):(got-want)));
}

static void updatePoints(int hand, int player) {
    points[hand][player] = calcNewPoints(
        prefs.hands[hand][player].want,
        prefs.hands[hand][player].got
    );
    if ((hand > 0) && (EMPTY != points[hand][player]))
        points[hand][player] += points[hand-1][player];
}

static void restoreAllPoints() {
    for (int h=0; h<MaxHands; ++h)
        for (int p=0; p<MaxPlayers; ++p)
            updatePoints(h, p);
}

static int countPlayers () {
    int p = 0;
    for (p=0; p<MaxPlayers; ++p)
        if (!StrCompare("", &prefs.players[p]))
            break;
    numPlayers = p;
    return numPlayers;
}

static void initGame() {
    firstVisibHand = 0;
    currHand = 0;
    currPlayer = 0;
    stage = SET;
    updateHands();
    clearAllPoints();
}

static void debugGame() {
    char hS[4], pS[4], sS[2];
    StrIToA(&hS, currHand);
    StrIToA(&pS, currPlayer);
    sS[1]='\0';
    switch (stage) {
        case SET: sS[0]='S'; break;
        case WANT: sS[0]='W'; break;
        case PLAY: sS[0]='P'; break;
        case GOT: sS[0]='G'; break;
        case REST: sS[0]='R'; break;
    }
    FrmCustomAlert(DebugAlert,
        &hS, &pS, &sS);
}

static void restoreGame() {
    firstVisibHand = 0;
    stage = SET;
    // game not started
    if (MinPlayers > countPlayers())
        return;
    // game started
    updateHands();
    restoreAllPoints();
    // find current hand, player, stage
    int h, p;
    for(h = 0; h < numHands; ++h) {
        // want
        for(p = 0; p < numPlayers; ++p)
            if (EMPTY == prefs.hands[h][actualPlayer(h, p)].want) {
                stage = WANT;
                break;
            }
        if (SET != stage)
            break;
        // got
        for(p = 0; p < numPlayers; ++p)
            if (EMPTY == prefs.hands[h][actualPlayer(h, p)].got) {
                stage = GOT;
                break;
            }
        if (SET != stage)
            break;
    }
    currHand = h;
    currPlayer = p;
    if (currHand == numHands) {
        --currHand;
        stage = REST;
    }
    if (0 == currPlayer) {
        if (WANT == stage) {
            --currHand;
            stage = REST;
        }
        else
            stage = PLAY;
    }
}

static void * getObjPtr(FormPtr frmP, int b) {
    return FrmGetObjectPtr(frmP,
        FrmGetObjectIndex(frmP, b));
}

static void drawGrid() {
    // vertical lines
    for (int p=0; p<numPlayers+1; ++p)
        WinDrawLine(
            HW + pW*p, TY,
            HW + pW*p, TY + PH + HH*VisibHands);
    // horizontal lines
    for (int h=0; h<VisibHands+1; ++h)
       WinDrawLine(
           TX, TY + PH + HH*h,
           HW + pW*numPlayers,TY + PH + HH*h);
    // players
    FntSetFont(boldFont);
    for (int p=0; p<numPlayers; ++p) {
        int len = StrLen(&prefs.players[p]);
        int width = pW - 2;
        Boolean fits;
        FntCharsInWidth(&prefs.players[p],
            &width, &len, &fits);
        WinDrawChars(&prefs.players[p], len,
            HW + pW*p + 2, TY + 1);
    }
}

static void drawPoints() {
    int maxVisibHands = (numHands > VisibHands)
        ? VisibHands : numHands;
    for (int h=0; h<maxVisibHands; ++h) {
        int aH = h + firstVisibHand; // actual hand
        // background
        RectangleType r;
        RctSetRectangle(&r,
            0, TY + PH + HH*h + 1,
            HW + pW*numPlayers, HH - 1);
        WinSetBackColor(WinRGBToIndex(
            (aH % 2) ? &rowBackOdd : &rowBackEven));
        WinEraseRectangle(&r, 0);
        // hand
        char hS = '0' + roundsPerHand[aH];
        FntSetFont(largeFont);
        WinDrawChars(&hS, 1,
            TX + 3, TY + PH + HH*h + 8);
        // players
        for (int p=0; p<numPlayers; ++p) {
            char wS, gS, aS[3];
            int w = prefs.hands[aH][p].want;
            int g = prefs.hands[aH][p].got;
            int a = points[aH][p];
            // want
            if (EMPTY != w) {
                wS = '0' + w;
                FntSetFont(actualPlayer(aH, 0) == p
                    ? boldFont : stdFont);
                WinDrawChars(&wS, 1,
                    HW + pW*p + 2, TY + PH + HH*h + 1);
            }
            // got
            if (EMPTY != g) {
                gS = '0' + g;
                FntSetFont(stdFont);
                WinDrawChars(&gS, 1,
                    HW + pW*p + pW - 6, TY + PH + HH*h + 1);
            }
            // points
            if (EMPTY != a) {
                StrIToA(&aS, a);
                FntSetFont(boldFont);
                int len = StrLen(&aS);
                int width = FntCharsWidth(&aS, len);
                WinDrawChars(&aS, len,
                   HW + pW*p + ((pW-width)/2),
                   TY + PH + HH*h + 11);
            }
            // vertical line
            WinDrawLine(
                HW + pW*p, TY + PH + HH*h,
                HW + pW*p, TY + PH + HH*h + HH);
        }
    }
}

static void drawScrollbar(FormPtr frmP) {
    int max = (numHands > VisibHands)
        ? (numHands - VisibHands) : 0;
    ScrollBarPtr barP = getObjPtr(frmP, HandScrollBar);
    SclSetScrollBar(barP, firstVisibHand,
        0, max, VisibHands);
}

static Err nameLoadItem(VoidPtr table, Word row, Word  column,
    Boolean editable, VoidHand  * dataH, WordPtr dataOffset,
    WordPtr dataSize, FieldPtr fld)
{
    //FieldAttrType attr;
    //FldGetAttributes(getObjPtr(FrmGetActiveForm(), SampleNameField),
    //    &attr);
    //attr.autoShift = true;
    //FldSetAttributes(fld, &attr);
    FldSetUsable(fld, true);
    FldSetMaxChars(fld, MaxNameLen);

    *dataH = nameHandles[row];
    StrCopy(MemHandleLock(*dataH),
        &prefs.players[row]);
    MemHandleUnlock(*dataH);
    *dataOffset = 0;
    *dataSize = MemHandleSize(* dataH);
    return 0;
};

static Boolean nameSaveItem(VoidPtr table,
    Word row, Word column)
{
    Boolean result = false;
    //FieldPtr field =  TblGetCurrentField(table);
    //if (field &&  FldDirty(field)) {
        VoidHand h =  nameHandles[row];
        CharPtr s = MemHandleLock(h);
        // uppercase first char
        if ((s[0] >= 'a') &&  (s[0] <= 'z'))
            s[0] += 'A' - 'a';
        StrCopy(&prefs.players[row], s);
        MemHandleUnlock(h);
        TblMarkRowInvalid(table, row);
        result = true;
    //}
    return result;  // should  the table be redrawn
}

static void mainFormInit(FormPtr frmP) {
    // buttons
    if (REST == stage) {
        if (-1 == currHand)
            CtlHideControl(getObjPtr(frmP,
                BackButton));
        if ((numHands-1) <= currHand)
            CtlHideControl(getObjPtr(frmP,
                NextButton));
    }
    // grid
    FrmDrawForm(frmP);
    pW = (PW * MaxPlayers) / numPlayers;
    drawGrid();
    drawPoints();
    drawScrollbar(frmP);
}

static void playerFormInit(FormPtr frmP) {
    // players
    TablePtr tableP = FrmGetObjectPtr(frmP,
        FrmGetObjectIndex(frmP, PlayerTable));

    for (int i = 0; i < MaxPlayers; ++i) {
        TblSetItemStyle(tableP,
            i, 0, numericTableItem);
        TblSetItemInt(tableP, i, 0, i+1);
        TblSetItemStyle(tableP,
            i, 1, textTableItem);
        TblSetRowUsable(tableP, i, true);
        TblSetRowSelectable(tableP, i, true);
    }

    TblSetColumnWidth(tableP, 0, 10);
    TblSetColumnWidth(tableP, 1, MaxNameLen*6);
    TblSetColumnUsable(tableP, 0, true);
    TblSetColumnUsable(tableP, 1, true);
    TblSetLoadDataProcedure(tableP,
        1, nameLoadItem);
    TblSetSaveDataProcedure(tableP,
        1, nameSaveItem);

    // start rounds
    CtlSetValue(getObjPtr(frmP,
        RoundsBaseButton + prefs.startRounds), true);

    FrmDrawForm(frmP);
}

static void wantFormInit(FormPtr frmP) {
    // actual player
    int actual = actualPlayer(currHand, currPlayer);
    // player name
    CtlSetLabel(getObjPtr(frmP, RoundsNameLabel),
        prefs.players[actual]);
    // accumulated
    int acum = 0;
    for (int p=0; p<currPlayer; ++p)
        acum += prefs.hands[currHand][actualPlayer(currHand, p)].want;
    char *acumS = CtlGetLabel(getObjPtr(frmP,
        RoundsAcumLabel));
    StrIToA(acumS, acum);
    // current rounds
    int rounds = prefs.hands[currHand][actual].want;
    Boolean nextUsable = false;
    if (EMPTY != rounds) {
        // preselect existing
        CtlSetValue(getObjPtr(frmP,
            RoundsBaseButton + rounds), true);
        nextUsable = true;
    }

    // hide bigger than curr. rounds
    for (int r= roundsPerHand[currHand]+1; r <= MaxRounds; ++r)
        CtlHideControl(getObjPtr(frmP,
            RoundsBaseButton + r));
    // hide invalid option for last player
    if ( ((currPlayer+1) == numPlayers)
        && (acum <= roundsPerHand[currHand]) )
    {
        int invalid = roundsPerHand[currHand] - acum;
        CtlHideControl(getObjPtr(frmP,
            RoundsBaseButton + invalid));
        if (invalid == rounds)
            nextUsable = false;
    }

    // can continue
    if (nextUsable)
        CtlShowControl(getObjPtr(frmP,
            NextButton));

    FrmDrawForm(frmP);
}

static void gotFormInit(FormPtr frmP) {
    // actual player
    int actual = actualPlayer(currHand, currPlayer);
    // player name
    CtlSetLabel(getObjPtr(frmP, RoundsNameLabel),
        prefs.players[actual]);
    // accumulated
    int acum = 0;
    for (int p=0; p<currPlayer; ++p)
        acum += prefs.hands[currHand][actualPlayer(currHand, p)].got;
    // want
    char *wantS = CtlGetLabel(getObjPtr(frmP,
        RoundsWantLabel));
    StrIToA(wantS, prefs.hands[currHand][actual].want);
    // current rounds
    int rounds = prefs.hands[currHand][actual].got;
    Boolean nextUsable = false;
    if (EMPTY != rounds) {
        // preselect existing
        CtlSetValue(getObjPtr(frmP,
            RoundsBaseButton + rounds), true);
        nextUsable = true;
    }

    // hide bigger than curr. rounds
    for (int r= roundsPerHand[currHand]+1; r <= MaxRounds; ++r)
        CtlHideControl(getObjPtr(frmP,
            RoundsBaseButton + r));
    // hide if will exceed
    for (int r=(roundsPerHand[currHand] - acum + 1); r<=roundsPerHand[currHand]; ++r) {
        CtlHideControl(getObjPtr(frmP,
            RoundsBaseButton + r));
        if (r == rounds)
            nextUsable = false;
    }
    // show single valid option for last player    
    if ( ((currPlayer+1) == numPlayers)
        && (acum <= roundsPerHand[currHand]) )
    {
        for (int r=0; r<=roundsPerHand[currHand]; ++r)
            CtlHideControl(getObjPtr(frmP,
                RoundsBaseButton + r));
        int valid = roundsPerHand[currHand] - acum;
        CtlShowControl(getObjPtr(frmP,
            RoundsBaseButton + valid));
        if (valid != rounds)
            nextUsable = false;
    }

    // can continue
    if (nextUsable)
        CtlShowControl(getObjPtr(frmP,
            NextButton));

    FrmDrawForm(frmP);
}

static void switchForm() {
    //debugGame();
    switch (stage) {
        case SET:
            FrmGotoForm(PlayerForm);
            break;
        case WANT:
             FrmGotoForm(WantForm);
            break;
        case GOT:
             FrmGotoForm(GotForm);
            break;
        default: // PLAY, REST
             FrmGotoForm(MainForm);
     }
}

static void drawInt(int n, int x, int y, int cw) {
    char t[8];
    StrIToA(&t, n);
    int l = StrLen(&t);
    int nw = FntCharsWidth(&t, l);
    WinDrawChars(&t, l, x + cw - nw, y);
}

static int roundRational(int n, int d) {
    return (
        (n >= 0 ? 5 : -5) +
        ((n*10) / d)
    ) / 10;
}

static void showStats() {
    // last completed hand
    int lastHand = (REST == stage)
        ? currHand : currHand-1;
    // order players by points
    int position[MaxPlayers];
    int handsWon[MaxPlayers];
    for (int p=0; p<numPlayers; ++p) {
        position[p] = p;
        handsWon[p] = 0;
        for (int h=0; h<=lastHand; ++h)
            if (prefs.hands[h][p].want == prefs.hands[h][p].got)
                ++handsWon[p];
    }
    Boolean ordered;
    do {
        ordered = true;
        for (int p=1; p<numPlayers; ++p)
            if (points[lastHand][position[p-1]] < points[lastHand][position[p]]) {
                int t = position[p];
                position[p] = position[p-1];
                position[p-1] = t;
                ordered = false;
            }
    } while (!ordered);
    // init form
    FormPtr frmP = FrmGetActiveForm();
    FormPtr alertP = FrmInitForm(
        StatsForm);
    FrmSetActiveForm(alertP);
    FrmDrawForm(alertP);
    // draw
    char mS[30];
    char nS[40];
    StrPrintF(&mS, FrmGetLabel(alertP, StatsHandsLabel),
        lastHand+1, numHands);
    FrmCopyLabel(alertP, StatsHandsLabel, &mS);
    FntSetFont(stdFont);
    for (int p=0; p<numPlayers; ++p) {
        int aP = position[p];
        int rY = ST_TY + (p+2)*ST_RH;
        StrPrintF(&nS, "%d. %s",
            p+1, &prefs.players[aP]);
        WinDrawChars(&nS, StrLen(&nS),
            ST_TX, rY);
        drawInt((lastHand >= 0) ? points[lastHand][aP] : 0,
            ST_TX + ST_NW,  rY, ST_SW);
        drawInt(handsWon[aP],
            ST_TX + ST_NW + ST_SW,  rY, ST_SW);
        drawInt(lastHand + 1 - handsWon[aP],
            ST_TX + ST_NW + ST_SW*2,  rY, ST_SW);
        drawInt((lastHand >= 0) ? roundRational(points[lastHand][aP], lastHand + 1) : 0,
            ST_TX + ST_NW + ST_SW*3,  rY, ST_SW);
    }
    // return
    FrmDoDialog(alertP);
    FrmDeleteForm(alertP);
    FrmSetActiveForm(frmP);
}

static Boolean doMenu(FormPtr frmP, Word command) {
    Boolean handled = false;
    switch (command) {
        case NewMenuOpt :
            if (0 == FrmAlert(NewGameAlert)) {
                initGame();
                switchForm();
            }
            handled = true;
            break;
        case StatsMenuOpt :
            showStats();
            handled = true;
            break;
        case AboutMenuOpt :
            FrmCustomAlert(AboutAlert,
                versionP, "", "");
            handled = true;
            break;
        default: // rules
            int rule = command - RulesBaseMenuOpt;
            if ((0 <= rule) && (rule < MaxRules)) {
                FrmHelp(RulesBaseStr + rule);
                handled = true;
            }
    }
    return handled;
}

static void dealAlert() {
    char rS[2];
    StrIToA(&rS, roundsPerHand[currHand]);
    FrmCustomAlert(DealAlert,
        prefs.players[actualPlayer(currHand, currPlayer-1)],
        &rS, "");
}

Boolean mainFormEventHandler(EventPtr eventP) {
    Boolean handled = false;
    FormPtr frmP = FrmGetActiveForm();
    switch (eventP->eType) {
        case frmOpenEvent:
            mainFormInit(frmP);
            handled = true;
            break;
        case keyDownEvent:
            char c = eventP->data.keyDown.chr;
            if ((pageUpChr == c)
                && (firstVisibHand > 0))
            {
                --firstVisibHand;
                drawPoints();
                drawScrollbar(frmP);
                handled = true;
            }
            if ((pageDownChr == c)
                && ((numHands - firstVisibHand) > VisibHands))
            {
                ++firstVisibHand;
                drawPoints();
                drawScrollbar(frmP);
                handled = true;
            }
            break;                  
        case sclRepeatEvent:
            firstVisibHand = eventP->data.sclRepeat.newValue;
            drawPoints();
            handled = false; // !
            break;
        case ctlSelectEvent:
            int b = eventP->data.ctlSelect.controlID;
            if (NextButton == b) {
                if (PLAY == stage) {
                    currPlayer = 0;
                   stage = GOT;
               }
               else { // REST
                   ++currHand;
                   currPlayer = 0;
                   stage = WANT;
                   dealAlert();
               }
           }
           else if (BackButton == b) {
               currPlayer = numPlayers - 1;
               stage = (REST == stage)
                   ? GOT : WANT;
           }
            switchForm();
            handled = true;
            break;
    }
    return handled;
}

Boolean playerFormEventHandler(EventPtr eventP) {
    Boolean handled = false;
    FormPtr frmP = FrmGetActiveForm();
    switch (eventP->eType) {
        case frmOpenEvent:
            playerFormInit(frmP);
            handled = true;
            break;
        case ctlSelectEvent:
            int b = eventP->data.ctlSelect.controlID;
            int r = b - RoundsBaseButton;
            if ((MinRounds <= r) && (r <= MaxRounds))
                prefs.startRounds = r;
            else if (MinPlayers > countPlayers())
                FrmAlert(MinPlayersAlert);
            else {
                initGame();
                dealAlert();
                currPlayer = 0;
                stage = WANT;
                switchForm();
            }
            handled = true;
            break;
    }
    return handled;
}

Boolean wantFormEventHandler(EventPtr eventP) {
    Boolean handled = false;
    FormPtr frmP = FrmGetActiveForm();
    switch (eventP->eType) {
        case frmOpenEvent:
            wantFormInit(frmP);
            handled = true;
            break;
        case ctlSelectEvent:
            int b = eventP->data.ctlSelect.controlID;
            int r = b - RoundsBaseButton;
            if (BackButton == b) {
                if (0 == currPlayer) {
                    --currHand;
                    stage = REST;
                }
                else
                    --currPlayer;
            }
            if ((r>=0) && (r<=MaxRounds)) {
                int actual = actualPlayer(currHand, currPlayer);
                prefs.hands[currHand][actual].want = r;
            }
            if ( ((r>=0) && (r<=MaxRounds))
                || (NextButton == b) )
            {
                if (currPlayer < (numPlayers - 1))
                    ++currPlayer;
                else
                    stage = PLAY;
            }
            switchForm();
            handled = true;
            break;
    }
    return handled;
}

Boolean gotFormEventHandler(EventPtr eventP) {
    Boolean handled = false;
    FormPtr frmP = FrmGetActiveForm();
    switch (eventP->eType) {
        case frmOpenEvent:
            gotFormInit(frmP);
            handled = true;
            break;
        case ctlSelectEvent:
            int b = eventP->data.ctlSelect.controlID;
            int r = b - RoundsBaseButton;
            if (BackButton == b) {
                if (0 == currPlayer) {
                    stage = PLAY;
                }
                else
                    --currPlayer;
            }
            if ((r>=0) && (r<=MaxRounds)) {
                int actual = actualPlayer(currHand, currPlayer);
                prefs.hands[currHand][actual].got = r;
                updatePoints(currHand, actual);
            }
            if ( ((r>=0) && (r<=MaxRounds))
                || (NextButton == b) )
            {
                if (currPlayer < (numPlayers - 1))
                    ++currPlayer;
                else
                    stage = REST;
            }
            switchForm();
            handled = true;
            break;
    }
    return handled;
}

static void copyStrFromRes(Char *dst, ULong type, Int16 id) {
    MemHandle nH = DmGetResource(type, id);
    StrCopy(dst, (Char *) MemHandleLock(nH));
    MemHandleUnlock(nH);
    DmReleaseResource(nH);
}

void startApp() {
    // init name handles
    for (int p=0; p<MaxPlayers; ++p)
        nameHandles[p] = MemHandleNew(MaxNameLen+1);

    // get version
    copyStrFromRes(versionP, 'tver', VersionRes);

    // default initialization, since discovered Prefs was missing or old
    Word prefSize = sizeof(Prefs);
    if ((PrefGetAppPreferences (AppCreator, 1000, &prefs, &prefSize, true) == noPreferenceFound)
        || (prefSize != sizeof(Prefs)))
    {
        initGame();
        // start rounds
        prefs.startRounds = MinRounds;
        // empty names
        for (int p=0; p<MaxPlayers; ++p)
            StrCopy(&prefs.players[p], "");
    }
    // restore from Prefs
    else {
        restoreGame();
    }
}

void stopApp() {
    PrefSetAppPreferences(AppCreator,
        1000, 1, &prefs, sizeof(Prefs), true);
}

Boolean isCompatible() {
    UInt32 romVer;
    Err err = FtrGet(sysFtrCreator,
        sysFtrNumROMVersion, &romVer);
    return (romVer >= MinROMVersion);
}

Boolean appHandleEvent(EventPtr eventP) {
    FormPtr frmP;
    Int formId;
    Boolean handled = false;
    switch (eventP->eType) {
        case frmLoadEvent:
            formId = eventP->data.frmLoad.formID;
            frmP = FrmInitForm(formId);
            FrmSetActiveForm(frmP);
            if (formId == MainForm)
                FrmSetEventHandler(frmP, mainFormEventHandler);
            if (formId == PlayerForm)
                FrmSetEventHandler(frmP, playerFormEventHandler);
            if (formId == WantForm)
                FrmSetEventHandler(frmP, wantFormEventHandler);
            if (formId == GotForm)
                FrmSetEventHandler(frmP, gotFormEventHandler);
            handled = true;
            break;
        case menuEvent:
            handled = doMenu(frmP, eventP->data.menu.itemID);
            break;
    }
    return handled;
}

DWord PilotMain(Word cmd, char *cmdPBP, Word launchFlags) {
    EventType event;
    Word error;
    if (cmd == sysAppLaunchCmdNormalLaunch) {
        if (isCompatible()) {
            startApp();
            switchForm();
            do {
                EvtGetEvent(&event, evtWaitForever);
                 if (!SysHandleEvent (&event))
                 if (!MenuHandleEvent (0, &event, &error))
                 if (!appHandleEvent (&event))
                         FrmDispatchEvent (&event);
            } while (event.eType != appStopEvent);
            stopApp();
        }
        else {
            FrmAlert(IncompatAlert);
        }
        FrmCloseAllForms();
    }
    return 0;
}
